/* olainterface.cpp
   Santi Noreña 2013
   It includes two classes:
   olaWorker is the threading class that reads raw DMX from ola daemon and save it into a buffer ordered in layers
   olaInterface controls olaWorker and translates DMX values received from olaWorker into orders to RenderingManager and Source
*/
#include "olainterface.h"

olaInterface::olaInterface()
{
    qDebug() << tr("olaInterface|") << "Starting";
    m_thread = NULL;
    worker = new olaWorker();
    Q_CHECK_PTR(worker);
    readDataFromXML(); // Read the dmx setting from dmx.xml
    connect (worker, SIGNAL(dmx(int,int,int)), this, SLOT(dmx(int,int,int)), Qt::QueuedConnection); // The DMX values.
    initMediaLibrary(); // Init the media library reading the files in m_pathmedia
    m_msex = new msex(this); // Init the CITP/MSEx protocol
    Q_CHECK_PTR(m_msex);
    open();
}

olaInterface::~olaInterface()
{
    close();
}

// Starts olaWorker in a new thread.
void olaInterface::open()
{
     m_thread = new QThread;
     Q_CHECK_PTR(m_thread);
     worker->moveToThread(m_thread);
     connect(m_thread, SIGNAL(started()), worker, SLOT(olastart()));
     connect(m_thread, SIGNAL(finished()), m_thread, SLOT(deleteLater()));
     m_thread->start();
}

// Close olaWorker and finish the thread
void olaInterface::close()
{
    worker->blockSignals(true);
    worker->olastop();
    delete worker;
}

// Read the dmx settings for dmx.xml At the moment we need:
// - The path to the medias folder tree
// - The number of sources/layers controlled by DMX
// - The first DMX channel of each source/layer
// - The universe to bind in OLA

void olaInterface::readDataFromXML() {
    QFile* xmlFile = new QFile("dmx.xlm");
    if (!xmlFile->open(QIODevice::ReadOnly | QIODevice::Text)) {
        QMessageBox::critical(NULL,"Load XML File Problem",
                              "Couldn't open dmx.xml to load settings for olaInterface",
                              QMessageBox::Ok);
        return;
    }
    QXmlStreamReader* xmlReader = new QXmlStreamReader(xmlFile);
    int counter = 0;
    //Parse the XML until we reach end of it
    while(!xmlReader->atEnd() && !xmlReader->hasError()) {
        // Read next element
        QXmlStreamReader::TokenType token = xmlReader->readNext();
        //If token is just StartDocument - go to next
        if(token == QXmlStreamReader::StartDocument) {
            continue;
        }
        //If token is StartElement - read it
        if(token == QXmlStreamReader::StartElement) {
            if(xmlReader->name() == "dmxSettings") {
                int version = xmlReader->attributes().value("fileVersion").toLocal8Bit().toInt();
                if(version == 1) {
                    worker->m_layersNumber = xmlReader->attributes().value("layersNumber").toLocal8Bit().toInt();
                    m_pathmedia = xmlReader->attributes().value("path").toLocal8Bit();
                    continue;
                }
            }
            if (worker->m_layersNumber > MAX_SOURCE_DMX) {
                worker->m_layersNumber = MAX_SOURCE_DMX;
            }
            QString add = "layer";
            add.append(QString("%1").arg(counter));
            if((xmlReader->name() == add) && (counter < worker->m_layersNumber)) {
                dmxSettings temp;
                temp.address = xmlReader->attributes().value("dmx").toLocal8Bit().toInt() - 1;
                temp.universe = xmlReader->attributes().value("universe").toLocal8Bit().toInt();
                worker->m_settings.replace(counter ,temp);
                // If the universe is not in the list, append it.
                if(!worker->m_universe.contains(temp.universe)) {
                    worker->m_universe.append(temp.universe);
                }
            }
            counter++;
        }
    }
    if(xmlReader->hasError()) {
        QMessageBox::critical(NULL,"xmlFile.xml Parse Error",xmlReader->errorString(), QMessageBox::Ok);
    }
    //close reader and flush file
    xmlReader->clear();
    xmlFile->close();

    delete xmlReader;
    delete xmlFile;
}

////////////////////////////////////////////////////////////
//
// Parse new DMX
//
////////////////////////////////////////////////////////////

//New dmx. Connected with signal newdmx from olaInterface->Worker
void olaInterface::dmx(int layer, int channel, int value)
{
// This qDebug slows all the program. Uncomment only for debugging purpouse and comment again in normal use
//    qDebug() << tr("olaInterface|") << "newdmx layer" << layer << "channel" << channel << "value" << value;
    switch(channel){
    case DMX_FOLDER:// Folder / Text File
    case DMX_FILE:// File / Font File
        requestNewFile(layer, (worker->m_dmx[layer][DMX_SOURCETYPE] / 25));
        break;
    case DMX_SOURCETYPE:// Source type
        if(value < 25)
        {
            requestNewFile(layer, 0);
        } else if (value > 24 && value < 50) {  // Create a  video source
            requestNewFile(layer, 1);
        } else if (value > 49 && value < 75){   // Render a text
            requestNewFile(layer, 2);
        } else if (value > 74 && value < 100) { // Live Cam
            requestNewFile(layer, 3);
        }
        break;
    default:
        emit sendDmx(layer, channel, value);
        break;
    }
}

// Create a new media source
void olaInterface::requestNewFile(int layer, int type){
    int cooki = layer + 201;
    QString newfile = tr("%1 %2 ").arg(cooki).arg(type);
    int folder = worker->m_dmx[layer][DMX_FOLDER];
    int file = worker->m_dmx[layer][DMX_FILE];
    if (type == 1) {
        newfile.append("open ");
        // Select one mediafile from the media library
        if (folder < m_media->size()) {
            if (file < m_media->at(folder).m_MediaInformation.size()) {
                newfile.append(m_media->at(folder).m_MediaInformation.at(file).MediaName);
                newfile.append(";");
                emit newFile(newfile);
            } else {
                qDebug() << tr("olaInterface|") << "Requested file is greater than files in library";
                return;
            }
        } else {
            qDebug() << tr("olaInterface|") << "Requested folder is greater than media libraries";
            return;
        }
    } else if (type == 2) {
        newfile.append("fuente font ");
        QDir dir;
        if (!dir.cd(m_pathmedia)) {
            qWarning() << "olaInterface::initMediaLibrary| Can not cd to the path: " << m_pathmedia;
            return;
        }
        if (!dir.cd("font")) {
            qWarning() << "olaInterface::initMediaLibrary| Can not cd to the media folder: " << m_pathmedia;
            return;
        }
        dir.setFilter(QDir::Files);
        QFileInfoList filelist = dir.entryInfoList();
        QFileInfo fileInfo = filelist.at(worker->m_dmx[layer][DMX_FILE]);
        newfile.append(fileInfo.absoluteFilePath().toAscii());
        newfile.append(";");
        emit newFile(newfile);
        newfile.clear();
        newfile = tr("%1 %2 ").arg(cooki).arg(type);
        newfile.append("text read ");
        dir.cdUp();
        if (!dir.cd("text")){
           qWarning() << "olaInterface::initMediaLibrary| Can not cd to text folder: " << m_pathmedia;
        }
        filelist = dir.entryInfoList();
        fileInfo = filelist.at(worker->m_dmx[layer][DMX_FOLDER]);
        newfile.append(fileInfo.absoluteFilePath().toAscii());
        newfile.append(";");
        emit newFile(newfile);
    } else if (type == 3) {

    }
}

// Initializes the media library and the media information from the path to media in m_pathmedia
void olaInterface::initMediaLibrary() {
    QDir dir;
    if (!dir.cd(m_pathmedia)) {
        qWarning() << "olaInterface::initMediaLibrary| Can not cd to the path: " << m_pathmedia;
        return;
    }
    if (!dir.cd("media")) {
        qWarning() << "olaInterface::initMediaLibrary| Can not cd to the media folder: " << m_pathmedia;
        return;
    }
    m_media = new QList<MediaLibrary>;
    dir.setFilter(QDir::Dirs | QDir::NoSymLinks | QDir::NoDotAndDotDot);
    QFileInfoList filelist = dir.entryInfoList();
    dir.setFilter(QDir::Files);
    QFileInfo fileInfo;
    struct MediaLibrary mediai;
    int i;
    for (i = 0; i < filelist.size(); ++i) {
        fileInfo = filelist.at(i);
        QString name = fileInfo.absoluteFilePath();
        dir.cd(fileInfo.baseName());
        mediai.m_Id = i;
        mediai.m_DMXRangeMax = i;
        mediai.m_DMXRangeMin = i;
        mediai.m_Name = name;
//        mediai.m_LibraryCount = 0; msex 1.2
        mediai.m_ElementCount = dir.count();
        mediai.m_MediaInformation = getMediaInformation(dir);
        dir.cdUp();
        m_media->append(mediai);
    }
}

// This set every media file included in one library/folder
QList<MediaInformation> olaInterface::getMediaInformation(QDir dir)
{
    QList<MediaInformation> mediaList;
    MediaInformation mediainf;
    dir.setFilter(QDir::Files);
    QFileInfoList filelist = dir.entryInfoList();
    QFileInfo fileInfo;
    bool folder = false;
    // Check if the thumbs folder exists and create if not
    if (!dir.cd("thumbs")) {
        if (!dir.mkdir("thumbs"))
        {
            qWarning() << tr("olaInterface|") << "Can not create thumbs folder in " << dir.path();
        } else {
        folder = true;
        dir.cd("thumbs");
        }
    }
    for (int i = 0; i < filelist.size(); ++i) {
        fileInfo = filelist.at(i);
        // Update the data base with the new file
        mediainf.Number = i;
        mediainf.DMXRangeMin = i;
        mediainf.DMXRangeMax = i;
        mediainf.MediaName = fileInfo.absoluteFilePath();
//        QDateTime date = fileInfo.created();
//        mediainf.MediaVersionTimestamp = date.toMSecsSinceEpoch();
        mediainf.MediaWidth = 320;
        mediainf.MediaHeight = 200;
        mediainf.MediaLength = 1000;
        mediainf.MediaFPS = 25;
        mediaList.append(mediainf);
        // Make the name of the thumb
        QString thumbname;
        thumbname.append(fileInfo.fileName());
        thumbname.append(".jpg");
        // Check if extis the thumbname. If not create it
        QStringList thumbList = dir.entryList();
        if ( !(thumbList.contains(thumbname)) || folder) { // Si no existe el fichero o hemos creado la carpeta gentera el thumbnail
            thumbname.prepend("/");
            thumbname.prepend(dir.path());
            QStringList arguments;
            arguments.append("-i");
            arguments.append(fileInfo.absoluteFilePath());
            arguments.append("-f");
            arguments.append("image2");
            arguments.append("-frames:v");
            arguments.append("1");
            arguments.append("-s");
            arguments.append("64x46");
            arguments.append("-vcodec");
            arguments.append("mjpeg");
            arguments.append(thumbname);
            QProcess avconv;
            // This should be done with ffmpeg library
            int fail = avconv.execute("avconv", arguments);
            if (fail != 0){
                qDebug() << tr("olaInterface|") << "avconv returned error" << fail;
            }
       }
    }
    return mediaList;
}

//////////////////////////////////////////////////
//
//          OLA Worker Class
//
//////////////////////////////////////////////////

    // --- CONSTRUCTOR ---
olaWorker::olaWorker() {
    m_clientpointer = NULL;
    m_counter = 0;
    gettimeofday(&m_last_data, NULL);
    m_layersNumber = MAX_SOURCE_DMX;
    dmxSettings newsetting;
    newsetting.address = -1;
    newsetting.universe = -1;
    for (int i=0; i < MAX_SOURCE_DMX; i++)
    {
        m_settings.append(newsetting);
        for (int j=0; j < LAYER_CHANNELS; j++)
        {
            m_dmx[i][j] = 0;
        }
    }
}

    // --- DECONSTRUCTOR ---
olaWorker::~olaWorker() {
    olastop();
}

   // Open the connection with olad and start processing data.
void olaWorker::olastart() {
    // set up ola connection
    m_client = new OlaCallbackClientWrapper;
    Q_CHECK_PTR(m_client);
    if (!m_client->Setup()) {qErrnoWarning("olaInterface::open: m_client.Setup:");}
    m_clientpointer = m_client->GetClient();
    m_clientpointer->SetDmxCallback(ola::NewCallback(this, &olaWorker::NewDmx));
    // register all the universe in list
    for (int i =0; i < m_universe.size(); i++) {
       m_clientpointer->RegisterUniverse(m_universe.at(i), ola::REGISTER,ola::NewSingleCallback(this, &olaWorker::RegisterComplete));
       qDebug() << tr("olaWorker|") << tr("Universe registered: %1").arg(m_universe.at(i));
    }
    m_client->GetSelectServer()->RegisterRepeatingTimeout(4000,ola::NewCallback(this, &olaWorker::CheckDataLoss));
    qDebug()<< tr("olaWorker|") << "Init complete.";
    m_client->GetSelectServer()->Run();
}

void olaWorker::olastop()
{
    if (m_clientpointer != NULL)
    {
        for (int i = 0; i < m_universe.size(); i ++) {
        m_clientpointer->RegisterUniverse(m_universe.at(i), ola::UNREGISTER, ola::NewSingleCallback(this, &olaWorker::RegisterComplete));
        }
        m_client->GetSelectServer()->Terminate();
        m_clientpointer = NULL;
    }
}

/*
 * CallBack from OLA daemon when there is new DMX data. This is called one for second if there is not updated in the DMX frame. We need emit only the channels that
 * has changed to save resources.
 */
// ToDo: It can be more efficient making the dmx buffer a DmxBuffer class instead a int array and compare with the new if there is changes at start. Also all access to the buffer it should be get/set. I should profile this
void olaWorker::NewDmx(unsigned int universe, const DmxBuffer &buffer,const string &error) {
    if (!error.empty()) {
        qWarning() << tr("olaWorker:NewDmx|") << "error received " << QString::fromStdString(error);
        return;
    }
    m_counter++;
    gettimeofday(&m_last_data, NULL);
    for (int i = 0; i < m_layersNumber; i++) { // loop for reading the channels by layer.
        if((m_settings.at(i).universe == universe) && ( m_settings.at(i).address > -1 )) { // Compare if the layer is from this universe
                                                                                         //  AND if the DMX address is 0 or greater, process this layer.
            for (int j = 0; j < LAYER_CHANNELS; j++){
                int value = buffer.Get((m_settings.at(i).address) + j); // Get the value for this channel.
                if (m_dmx[i][j] != value) { // Compare the new value with the old value.
                    emit dmx(i,j,value);  // Connected with dmx slot in olaInterface.
                    m_dmx[i][j] = value;
                }
            }
        }
    }
}

/*
 * Check for data loss each 4 seconds.
 */

bool olaWorker::CheckDataLoss() {
  struct timeval now, diff;
  if (timerisset(&m_last_data)) {
    gettimeofday(&now, NULL);
    timersub(&now, &m_last_data, &diff);
    if (diff.tv_sec > 4 || (diff.tv_sec == 4 && diff.tv_usec > 4000000)) {
      // loss of data
      qDebug()<< "olaWorker| Can not read one or several universes";
    }
  }
  return true;
}

/*
 * Control de errores en el registro de Universos en OLA
 */
void olaWorker::RegisterComplete(const string &error) {
  if (!error.empty()) {
      qWarning() << "olaWorker|" << "Register command failed" << QString::fromStdString(error);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////

/* We will need some GUI interaction from the user for settings. We will need:
 * - Change the path to the medias
 * - Change the DMX address of each layer
 * - Change the universe to bind in ola
 * - Write and read different xml configuration files for different setups
 * - Start and stop the dmx reading

void savesetttingtoxml{
// Habrá que salvar las settings en algún sitio
}

void opensettings{
Posibilidad de abrir otro archivo de configuración
}
*/
